Index: src/ia32/full-codegen-ia32.cc |
diff --git a/src/ia32/full-codegen-ia32.cc b/src/ia32/full-codegen-ia32.cc |
index 09a7aec9b8191618481a63b5bf59f87143ce9a5b..871179400378e66d39e25898e5bd0cd21a5e9c3d 100644 |
--- a/src/ia32/full-codegen-ia32.cc |
+++ b/src/ia32/full-codegen-ia32.cc |
@@ -218,21 +218,20 @@ void FullCodeGenerator::Generate(CompilationInfo* info) { |
Operand(ebp, StandardFrameConstants::kCallerSPOffset + offset)); |
__ push(edx); |
__ SafePush(Immediate(Smi::FromInt(scope()->num_parameters()))); |
- // Arguments to ArgumentsAccessStub and/or New...: |
+ // Arguments to ArgumentsAccessStub: |
// function, receiver address, parameter count. |
// The stub will rewrite receiver and parameter count if the previous |
// stack frame was an arguments adapter frame. |
- ArgumentsAccessStub::Type type; |
- if (is_strict_mode()) { |
- type = ArgumentsAccessStub::NEW_STRICT; |
- } else if (function()->has_duplicate_parameters()) { |
- type = ArgumentsAccessStub::NEW_NON_STRICT_SLOW; |
- } else { |
- type = ArgumentsAccessStub::NEW_NON_STRICT_FAST; |
- } |
- ArgumentsAccessStub stub(type); |
+ ArgumentsAccessStub stub( |
+ is_strict_mode() ? ArgumentsAccessStub::NEW_STRICT |
+ : ArgumentsAccessStub::NEW_NON_STRICT); |
__ CallStub(&stub); |
+ Variable* arguments_shadow = scope()->arguments_shadow(); |
+ if (arguments_shadow != NULL) { |
+ __ mov(ecx, eax); // Duplicate result. |
+ Move(arguments_shadow->AsSlot(), ecx, ebx, edx); |
+ } |
Move(arguments->AsSlot(), eax, ebx, edx); |
} |
@@ -375,7 +374,7 @@ void FullCodeGenerator::TestContext::Plug(Slot* slot) const { |
// For simplicity we always test the accumulator register. |
codegen()->Move(result_register(), slot); |
codegen()->PrepareForBailoutBeforeSplit(TOS_REG, false, NULL, NULL); |
- codegen()->DoTest(true_label_, false_label_, fall_through_); |
+ codegen()->DoTest(this); |
} |
@@ -449,7 +448,7 @@ void FullCodeGenerator::TestContext::Plug(Handle<Object> lit) const { |
} else { |
// For simplicity we always test the accumulator register. |
__ mov(result_register(), lit); |
- codegen()->DoTest(true_label_, false_label_, fall_through_); |
+ codegen()->DoTest(this); |
} |
} |
@@ -485,7 +484,7 @@ void FullCodeGenerator::TestContext::DropAndPlug(int count, |
__ Drop(count); |
__ Move(result_register(), reg); |
codegen()->PrepareForBailoutBeforeSplit(TOS_REG, false, NULL, NULL); |
- codegen()->DoTest(true_label_, false_label_, fall_through_); |
+ codegen()->DoTest(this); |
} |
@@ -562,7 +561,8 @@ void FullCodeGenerator::TestContext::Plug(bool flag) const { |
} |
-void FullCodeGenerator::DoTest(Label* if_true, |
+void FullCodeGenerator::DoTest(Expression* condition, |
+ Label* if_true, |
Label* if_false, |
Label* fall_through) { |
ToBooleanStub stub; |
@@ -1209,12 +1209,13 @@ void FullCodeGenerator::EmitDynamicLoadFromSlotFastCase( |
void FullCodeGenerator::EmitVariableLoad(Variable* var) { |
// Three cases: non-this global variables, lookup slots, and all other |
// types of slots. |
Slot* slot = var->AsSlot(); |
- ASSERT((var->is_global() && !var->is_this()) == (slot == NULL)); |
+ Property* property = var->AsProperty(); |
- if (slot == NULL) { |
+ if (var->is_global() && !var->is_this()) { |
Comment cmnt(masm_, "Global variable"); |
// Use inline caching. Variable name is passed in ecx and the global |
// object on the stack. |
@@ -1224,7 +1225,7 @@ void FullCodeGenerator::EmitVariableLoad(Variable* var) { |
EmitCallIC(ic, RelocInfo::CODE_TARGET_CONTEXT, AstNode::kNoNumber); |
context()->Plug(eax); |
- } else if (slot->type() == Slot::LOOKUP) { |
+ } else if (slot != NULL && slot->type() == Slot::LOOKUP) { |
Label done, slow; |
// Generate code for loading from variables potentially shadowed |
@@ -1240,7 +1241,7 @@ void FullCodeGenerator::EmitVariableLoad(Variable* var) { |
context()->Plug(eax); |
- } else { |
+ } else if (slot != NULL) { |
Comment cmnt(masm_, (slot->type() == Slot::CONTEXT) |
? "Context slot" |
: "Stack slot"); |
@@ -1258,6 +1259,36 @@ void FullCodeGenerator::EmitVariableLoad(Variable* var) { |
} else { |
context()->Plug(slot); |
} |
+ |
+ } else { |
+ Comment cmnt(masm_, "Rewritten parameter"); |
+ ASSERT_NOT_NULL(property); |
+ // Rewritten parameter accesses are of the form "slot[literal]". |
+ |
+ // Assert that the object is in a slot. |
+ Variable* object_var = property->obj()->AsVariableProxy()->AsVariable(); |
+ ASSERT_NOT_NULL(object_var); |
+ Slot* object_slot = object_var->AsSlot(); |
+ ASSERT_NOT_NULL(object_slot); |
+ |
+ // Load the object. |
+ MemOperand object_loc = EmitSlotSearch(object_slot, eax); |
+ __ mov(edx, object_loc); |
+ |
+ // Assert that the key is a smi. |
+ Literal* key_literal = property->key()->AsLiteral(); |
+ ASSERT_NOT_NULL(key_literal); |
+ ASSERT(key_literal->handle()->IsSmi()); |
+ |
+ // Load the key. |
+ __ SafeSet(eax, Immediate(key_literal->handle())); |
+ |
+ // Do a keyed property load. |
+ Handle<Code> ic = isolate()->builtins()->KeyedLoadIC_Initialize(); |
+ EmitCallIC(ic, RelocInfo::CODE_TARGET, GetPropertyId(property)); |
+ |
+ // Drop key and object left on the stack by IC. |
+ context()->Plug(eax); |
} |
} |
@@ -1490,7 +1521,7 @@ void FullCodeGenerator::VisitAssignment(Assignment* expr) { |
} |
// Left-hand side can only be a property, a global or a (parameter or local) |
- // slot. |
+ // slot. Variables with rewrite to .arguments are treated as KEYED_PROPERTY. |
enum LhsKind { VARIABLE, NAMED_PROPERTY, KEYED_PROPERTY }; |
LhsKind assign_type = VARIABLE; |
Property* property = expr->target()->AsProperty(); |
@@ -1516,13 +1547,29 @@ void FullCodeGenerator::VisitAssignment(Assignment* expr) { |
break; |
case KEYED_PROPERTY: { |
if (expr->is_compound()) { |
- VisitForStackValue(property->obj()); |
- VisitForAccumulatorValue(property->key()); |
+ if (property->is_arguments_access()) { |
+ VariableProxy* obj_proxy = property->obj()->AsVariableProxy(); |
+ MemOperand slot_operand = |
+ EmitSlotSearch(obj_proxy->var()->AsSlot(), ecx); |
+ __ push(slot_operand); |
+ __ SafeSet(eax, Immediate(property->key()->AsLiteral()->handle())); |
+ } else { |
+ VisitForStackValue(property->obj()); |
+ VisitForAccumulatorValue(property->key()); |
+ } |
__ mov(edx, Operand(esp, 0)); |
__ push(eax); |
} else { |
- VisitForStackValue(property->obj()); |
- VisitForStackValue(property->key()); |
+ if (property->is_arguments_access()) { |
+ VariableProxy* obj_proxy = property->obj()->AsVariableProxy(); |
+ MemOperand slot_operand = |
+ EmitSlotSearch(obj_proxy->var()->AsSlot(), ecx); |
+ __ push(slot_operand); |
+ __ SafePush(Immediate(property->key()->AsLiteral()->handle())); |
+ } else { |
+ VisitForStackValue(property->obj()); |
+ VisitForStackValue(property->key()); |
+ } |
} |
break; |
} |
@@ -1726,7 +1773,7 @@ void FullCodeGenerator::EmitAssignment(Expression* expr, int bailout_ast_id) { |
} |
// Left-hand side can only be a property, a global or a (parameter or local) |
- // slot. |
+ // slot. Variables with rewrite to .arguments are treated as KEYED_PROPERTY. |
enum LhsKind { VARIABLE, NAMED_PROPERTY, KEYED_PROPERTY }; |
LhsKind assign_type = VARIABLE; |
Property* prop = expr->AsProperty(); |
@@ -1786,6 +1833,8 @@ void FullCodeGenerator::EmitAssignment(Expression* expr, int bailout_ast_id) { |
void FullCodeGenerator::EmitVariableAssignment(Variable* var, |
Token::Value op) { |
+ // Left-hand sides that rewrite to explicit property accesses do not reach |
+ // here. |
ASSERT(var != NULL); |
ASSERT(var->is_global() || var->AsSlot() != NULL); |
@@ -3738,7 +3787,7 @@ void FullCodeGenerator::VisitCountOperation(CountOperation* expr) { |
} |
// Expression can only be a property, a global or a (parameter or local) |
- // slot. |
+ // slot. Variables with rewrite to .arguments are treated as KEYED_PROPERTY. |
enum LhsKind { VARIABLE, NAMED_PROPERTY, KEYED_PROPERTY }; |
LhsKind assign_type = VARIABLE; |
Property* prop = expr->expression()->AsProperty(); |
@@ -3765,8 +3814,16 @@ void FullCodeGenerator::VisitCountOperation(CountOperation* expr) { |
__ push(eax); |
EmitNamedPropertyLoad(prop); |
} else { |
- VisitForStackValue(prop->obj()); |
- VisitForAccumulatorValue(prop->key()); |
+ if (prop->is_arguments_access()) { |
+ VariableProxy* obj_proxy = prop->obj()->AsVariableProxy(); |
+ MemOperand slot_operand = |
+ EmitSlotSearch(obj_proxy->var()->AsSlot(), ecx); |
+ __ push(slot_operand); |
+ __ SafeSet(eax, Immediate(prop->key()->AsLiteral()->handle())); |
+ } else { |
+ VisitForStackValue(prop->obj()); |
+ VisitForAccumulatorValue(prop->key()); |
+ } |
__ mov(edx, Operand(esp, 0)); |
__ push(eax); |
EmitKeyedPropertyLoad(prop); |
@@ -1209,12 +1209,13 @@ void FullCodeGenerator::EmitDynamicLoadFromSlotFastCase( |
void FullCodeGenerator::EmitVariableLoad(Variable* var) { |
- // Three cases: non-this global variables, lookup slots, and all other |
- // types of slots. |
+ // Four cases: non-this global variables, lookup slots, all other |
+ // types of slots, and parameters that rewrite to explicit property |
+ // accesses on the arguments object. |
Slot* slot = var->AsSlot(); |